structure Parser :> PARSER =
  struct
    type filename = string

    local 
      fun ++ (r as (ref z)) = (r := (z + 1); z + 1)
      val p = ref 0
    in
      fun newpackid () = ++ p
    end
  
    local 
      val lc = ref 0
    in
      fun lcreset () = lc := 0
      fun lcn () = Int.toString (!lc)
      fun discardcomments cline =
        let
          val sl = String.tokens (fn x => Char.compare (x, #";") = EQUAL) cline
        in
          lc := !lc + 1;
          if List.null sl orelse String.compare (String.extract (cline, 0, SOME 1), ";") = EQUAL
            then NONE
          else SOME (List.hd sl)
        end
    end

    exception FoundLine of string
    exception EOF

    fun pr l = TextIO.output (TextIO.stdErr, String.concat l)
    val stoi = Option.valOf o Int.fromString

    val blank = List.all Char.isSpace o String.explode

    fun read1line file =
      ((while not (TextIO.endOfStream file) do
          (case discardcomments (TextIO.inputLine file) of
             NONE => ()
           | SOME l => if blank l then () else raise FoundLine l)); "")
          handle FoundLine l => l (* (pr ["line: \"", l, "\"\n"]; l) *)

    fun makepack s : Common.package =
      let
        val px :: py :: dx :: dy :: wei :: _ = String.tokens Char.isSpace s
        val [px, py, dx, dy, wei] =
          List.map (Option.valOf o Int.fromString) [px, py, dx, dy, wei]
      in
        (newpackid (),                  (* packid ??? *)
         wei,                           (* weight *)
         SOME (px, py),                 (* src *)
         (dx, dy))                      (* dest *)
      end

    fun spotconv #"~" = SOME Common.WATER
      | spotconv #"." = SOME Common.EMPTY
      | spotconv #"*" = SOME Common.EMPTY (* base *)
      | spotconv #"#" = SOME Common.WALL
      | spotconv #"@" = SOME Common.BASE
      | spotconv c = NONE

    val makespots = 
      Array.fromList o List.concat o List.map (List.map Option.valOf o List.filter Option.isSome o
                                               List.map spotconv o String.explode)
      (* if that composition seems obvious to you, you are indeed a leet ml master *)

    fun makespawn s =
      let
        val spawnx :: spawny :: _ = String.tokens Char.isSpace s
      in
        (stoi spawnx, stoi spawny)
      end

    datatype file = FILE of {world : Types.world,
                             spawns : Common.loc list}

    fun parse filename : file =
      let
        val is = TextIO.openIn filename
        val _ = lcreset ()
        val width :: height :: _ = String.tokens Char.isSpace (read1line is)
        val (width, height) = (stoi width + 2, stoi height + 2)
        val spots = 
          (String.implode (List.tabulate (width, fn _ => #"#"))
           :: List.tabulate (height - 2, fn _ => "#" ^ read1line is ^ "#"))
          @ [String.implode (List.tabulate (width, fn _ => #"#"))]
        val npacks = (stoi o read1line) is
        val packages = List.tabulate (npacks, fn _ => read1line is)
        val nspawns = (stoi o read1line) is
        val spawns = List.tabulate (nspawns, fn _ => (makespawn o read1line) is)
      in
        FILE {world = {board = makespots spots,
                       packages = List.map makepack packages,
                       robots = [],
                       cap = 0,
                       width = width,
                       height = height},
              spawns = spawns}
      end handle e => (TextIO.output (TextIO.stdErr, String.concat [filename, ":", lcn (), ":",
                                                                    exnMessage e, "\n"]); raise e)
  end
