
(* mutable priority queues indexed by integers *)

structure PQueue :> PQUEUE =
struct
    (* max frontier would be sqrt(2 million), so say 4000 to be safe *)
    val MAX = 10000

    type pqueue = (int Array.array *
		   (int * int) Array.array *
		   int ref)

    (* array is 1-based, this makes math easier (!) *)

    fun empty () = (Array.array (MAX, 0),
		    Array.array (MAX, (0,0)),
		    ref 1)

    fun show (ps, vs : (int * int) Array.array, ref l) =
	let
	in
	    print ("queue has idx " ^ Int.toString l ^ "\n");
	    Util.for 1 (l - 1)
	    (fn j =>
	     print ("  " ^ Int.toString j ^ ": " ^ Int.toString (Array.sub(ps, j)) ^ " = "
		    ^ Int.toString (#1 (Array.sub(vs, j))) ^ ", "
		    ^ Int.toString (#2 (Array.sub(vs, j))) ^ "\n"))
	end


    (* if node is smaller than its parent, swap and recurse *)
    fun perc_up (ps, vs, 1) = () (* root *)
      | perc_up (ps, vs, idx) =
	let val parp = Array.sub(ps, idx div 2)
	    val p = Array.sub(ps, idx)
	in
	    (* print (Int.toString idx ^ " perc up ...\n"); *)
	    if parp > p then
		let
		    val parv = Array.sub(vs, idx div 2)
		    val v = Array.sub(vs, idx)
		in
		    Array.update(vs, idx div 2, v);
		    Array.update(vs, idx, parv);
		    Array.update(ps, idx div 2, p);
		    Array.update(ps, idx, parp);
		    perc_up (ps, vs, idx div 2)
		end
	    else ()
	end

    fun insert (ps, vs, r as ref idx) p v =
 	let in
	    if idx >= MAX then 
		print "Out of space!!!\n"
	    else
		(Array.update(ps, idx, p);
		 Array.update(vs, idx, v);
		 r := idx + 1;
		 (*
		  print "bef perc:\n";
		  show (ps, vs, r);
		  *)
		 perc_up (ps, vs, idx);
	         (*
		  print "after perc:\n";
		  show (ps, vs, r);
		  *)
		 ())
	end

    fun perc_down (ps, vs, last, i) =
	let 
	    val li = i * 2
	    val ri = li + 1
	in
	    (* print (" perc_down(" ^ Int.toString last ^ " " ^ Int.toString i ^ ")...\n"); *)
	    if li >= last then ((* print " .. no children\n" *))
	    else if ri < last then
		(* two children. swap with smaller. *)
		let
		    val parv = Array.sub(vs, i)
		    val parp = Array.sub(ps, i)
		    val rv = Array.sub(vs, ri)
		    val rp = Array.sub(ps, ri)
		    val lv = Array.sub(vs, li)
		    val lp = Array.sub(ps, li)
		in
		    if parp < lp andalso parp < rp then ((*print " .. children larger\n"*))
		    else if lp < rp then
			    (* swap L child *)
			    let in
				(* print (Int.toString i ^ " swap l...\n"); *)
				Array.update(ps, i, lp);
				Array.update(vs, i, lv);
				Array.update(ps, li, parp);
				Array.update(vs, li, parv);
				perc_down (ps, vs, last, li)
			    end
			 else
			     (* swap R *)
			    let in
				(* print (Int.toString i ^ " swap r...\n"); *)
				Array.update(ps, i, rp);
				Array.update(vs, i, rv);
				Array.update(ps, ri, parp);
				Array.update(vs, ri, parv);
				perc_down (ps, vs, last, ri)
			    end
		end
		 else (* otherwise one child:
		         perc down parent = perc up for child *)
		     let in
			 (* print (Int.toString i ^ " use perc_up...\n"); *)
			 perc_up (ps, vs, li)
		     end
	end

    fun remlowest (_, _, ref 1) = NONE
      | remlowest (ps, vs, r as ref idx) =
	let
	    (* root will be returned *)
	    val ret = Array.sub(vs, 1)
	    (* put last element in root pos *)
	    val v = Array.sub(vs, idx - 1)
	    val p = Array.sub(ps, idx - 1)
	in
	    r := idx - 1;
	    Array.update(ps, 1, p);
	    Array.update(vs, 1, v);
	    perc_down (ps, vs, idx - 1, 1);
	    SOME ret
	end

end