
structure StupidBotIV :> ROBOT =
struct
    open Comm
    open World

    (* go to r, c, deliver pid *)
    datatype plan = 
        GOTO of (int * int) * int option

    val showpacks = Params.flag false (SOME ("-showpacks", 
					     "Show the known packages at every turn.")) "showpacks"

    val debug = Params.flag false (SOME ("-debug", 
					 "Turn on debugging messages.")) "debug"

    type mind = plan

    fun dist ((r,c),(r',c')) = Int.abs(r-r') + Int.abs(c-c')

    (* find a base to GOTO (except for the one listed) *)
    fun goto_base_not (a,b) (board as BOARD{layout, cap, ...}) (me as ROBOT{loc, ...}) =
	case RobUtil.findspots board (fn (r, c, SPOT{space=Dot true,...}) => (r <> a orelse b <> c) | _ => false) of
	    nil => 
		(case RobUtil.findspot board (fn (r, c, SPOT{space=Dot _,...}) => 
					      (r <> a orelse b <> c) | _ => false) of
		   SOME l => GOTO (l,NONE)
                 | NONE => GOTO ((1,1),NONE)) (* ??? *)
	  | h::t =>
            let in
		GOTO (foldl (fn (loc',best) => if dist (loc,loc') < dist(loc,best) then loc' else best) h t, NONE)
            end
					     

    val init = goto_base_not (~1, ~1)
    val itos = Int.toString
    fun dtos (a,b) = itos a ^ "," ^ itos b
    fun pltos (Ground d) = "@" ^ dtos d
      | pltos (Robot r) = "[rob " ^ itos r ^ "]"

    fun ptos (PACKAGE{id,loc,weight,dest}) = itos id

    fun ktos f (Known n) = f n
      | ktos _ _ = "Unknown"

    fun move (board as BOARD{layout, packages, robots, cap}) dmap (GOTO((a,b),deliver))
	                    (me as ROBOT{loc as (r, c), packages=packs, money, ...}) =
	let 

	    val adjacent = [(~1, 0), (1, 0), (0, 1), (0, ~1)]
	    val numrobs = foldl (fn ((dr,dc), n) =>
				 case sub layout (r + dr, c + dc) of
				     SPOT{robots, ...} => n + length robots) 0 adjacent

	    (* if there are any robots, also add up the amount of surrounding
	       water (danger!), especially if that water is opposite a robot!! *)
	    val water = if numrobs > 0 then foldl (fn ((dr,dc), n) => n + 
						   (case sub layout (r + dr, c + dc) of
						       SPOT{space = Water, ...} => 
							   (case sub layout (r - dr, c - dc) of
								SPOT{robots=_::_,...} => 25 (* ? *)
							      | _ => 2)
						     | _ => 0)) numrobs adjacent
			else 0

	    (* if a robot is next to me, bid 2 (or more ...!). *)
	    val bid = 1 + water
		
	    fun choose_deliver l = 
		let
		    val pl = map (RobUtil.getpackage packages) l
		    val pl = map (fn PACKAGE{dest = Known dest, id, ...} => 
				  ((id,dest), FloodFill.readloc dmap layout dest)
		  | PACKAGE{id=pid,...} => 
				  let in
				      print ("BAD PACKAGE !!! Held by me: " ^ 
					     itos pid ^ "\n");
				      raise Fail "oops" (* XXX protect *)
				  end) pl
		    val (((pid,dest), _)::_) = 
			ListUtil.sort (ListUtil.bysecond Int.compare) pl
		in
		    if !debug then print ("Now proceeding to " ^ dtos dest ^ " ...\n") else ();
			GOTO (dest, SOME pid)
		end

	in
	    if !debug andalso bid > 1 then print (" DANGER DETECTED OF LEVEL " ^ itos bid ^ "! \n") else ();

	    if !debug then
		print ("(StupidBot 4.0) I'm at " ^ itos r ^ "," ^ itos c ^ " heading to " ^ 
		       itos a ^ "," ^ itos b ^ 
		       (case deliver of NONE => " (just for fun)"
		        | SOME pid => " (to deliver " ^ itos pid ^ ")") ^
		       ". I have $" ^ itos money ^ ".\n")
	    else ();
	    if !showpacks then
	    print ("Location of known packages:\n     " ^
		   StringUtil.delimit "\n     " (map (fn PACKAGE{id, loc, weight, dest} =>
						      itos id ^ ": " ^ ktos dtos dest ^ " " ^ ktos itos weight ^ "lb. "
						      ^ ktos pltos loc) (World.Map.listItems packages)) ^ "\n")
	    else ();
	if bid < 25 andalso a = r andalso b = c then
	    (* got there! *)
	    (* pick up or put down? *)
	    (case RobUtil.packsdesthere loc packages packs of
		nil => (case packs of
                          nil => (* empty handed!! find a new base? *)
                            (case RobUtil.packsarehere loc packages of
                               nil =>                   
                                 let in
				     if !debug then print ("Why am I here? Finding a new base...\n") else ();
				     move board dmap (goto_base_not (a,b) board me) me
                                 end
                             | l => 
                                 let val l = ((Knapsack.pickup board layout dmap l (#1 cap)) handle _ => l) in
                                   if !debug then print ("Pick up: " ^ 
							 StringUtil.delimit " " (map ptos l) ^ "\n") else ();
                                   (GOTO((a,b), NONE), bid, Pick l)
                                 end)
                        | l => move board dmap (choose_deliver l) me)
              | l => 
                    let in
                        if !debug then print ("Deliver: " ^ StringUtil.delimit " " (map ptos l) ^ "\n") else ();
                        (GOTO((a,b), NONE), bid, Drop l)
                    end)
        else
            let 
		(* danger or not at destination -- execute move *)

		(* opportunistic kill? *)
		val kill = if numrobs > 0 then List.find (fn (dr,dc) =>
							  (case sub layout (r + dr, c + dc) of
							       SPOT{robots = _::_, ...} => 
								   (case sub layout (r + dr + dr , c + dc + dc) of
									SPOT{space = Water, ...} => true
								      | _ => false)
							  | _ => false)) adjacent
			    else NONE

		val (d, b, a) =
		 case kill of
		    NONE => let
				val _ = if !debug then print ("Find the fastest way to my destination ..\n") else ()
				val move = (Util.time "dirto" (FloodFill.dirto dmap board (a,b))) loc
				val _ = if !debug then print ("OK! got it! \n") else ()
			    in
				(GOTO((a,b), deliver), bid, move)
			    end
		  | SOME (1, 0) => (GOTO((a,b), deliver), bid * 4, Move N)
		  | SOME (~1, 0) => (GOTO((a,b), deliver), bid * 4, Move S)
		  | SOME (0, 1) => (GOTO((a,b), deliver), bid * 4, Move E)
		  | SOME _ => (GOTO((a,b), deliver), bid * 4, Move W)

		(* sanity check *)
		val b = if b >= money then 1 else b
		val d = (case d of
			     GOTO(_, NONE) => d
			   | GOTO(_, SOME p) =>
				 if List.exists (fn z => z = p) packs then d
				 else (if !debug then print "Trying to deliver a package I dropped!\n" else ();
				       choose_deliver packs))
	    (* also don't walk into water XXX *)
	    in
		(d, b, a)
            end
        end
end
