structure LSocket :> LSOCKET =
struct

    (* oh yeah *)
    type socket = int
    type listener = int

    exception Socket of string

    val pf_inet = _ffi "ML_PF_INET" : int ;
    val sock_stream = _ffi "ML_SOCK_STREAM" : int ;
    val msg_nosignal = _ffi "ML_MSG_NOSIGNAL" : SysWord.word ;

    (* keeps sending until exhausted *)
    fun sendall s x l =
	let
	    val sys_send = _ffi "send" : int * string * int * SysWord.word -> int ;
	in
	    case sys_send (s, x, l, msg_nosignal) of
		~1 => raise Socket "socket disconnected"
	      | n => if n < l then
		sendall s (CharVector.extract (x, n, NONE)) (l - n)
		else ()
	end

    fun send sk s = sendall sk (s ^ "\n") (size s + 1)
    fun sendraw sk s = sendall sk s (size s)

    fun recv sk =
	let
	    val sys_recv = _ffi "recv" : int * CharArray.array * int * SysWord.word -> int ;
	    val a = CharArray.array (1, #"?")

	    (* receive 1 char until newline *)
	    fun gc () = 
		(case sys_recv (sk, a, 1, msg_nosignal) of
		     1 => (case CharArray.sub (a, 0) of
			 #"\n" => ""
		       | #"\r" => (print "Warning! Got CR!\n"; gc ())
		       | c => implode[c] ^ gc())
		   | _ => raise Socket "socket error/disconnected")
	in
	    gc ()
	end

    fun hangup sk = 
	let
	    val cl = _ffi "close" : int -> unit ;
	in
	    cl sk
	end

    fun listen port = 
	let
	    val sys_socket = _ffi "socket" : int * int * int -> int ;
	    val sys_listen = _ffi "listen" : int * int -> int ;
	    val sys_bindport = _ffi "ml_bindport" : int * int -> int ;

	    val sk = sys_socket(pf_inet, sock_stream, 0)
	in
	    sys_bindport (sk, port) = 0 orelse raise Socket "can't bind";
	    sys_listen (sk, 20) = 0 orelse raise Socket "can't listen";
	    sk
	end

    fun accept l =
	let
	    val sys_accept = _ffi "ml_accept" : int -> int ;
	in
	    case sys_accept l of
		~1 => raise Socket "accept failed"
	      | s => s
	end

    fun connect a p =
	let
	    val sys_connect = _ffi "ml_connect" : string * int -> int ;
	in
	    case sys_connect (a ^ "\000", p) of
		~1 => raise Socket "connect failed"
	      | s => s
	end

end
